#!/usr/bin/env python3

from datetime import date
from csv import DictReader
from bz2file import BZ2File
from glob import glob
from itertools import chain
from sys import argv
from functional import compose, partial
from functools import reduce
import gflags
from glob import glob
from os.path import join, basename
from os import environ
from citibike import cache
from json import dumps
import sys
import os
FLAGS = gflags.FLAGS


gflags.DEFINE_string('glob', '*.csv.bz2', 'Glob pattern to read from cache')
gflags.DEFINE_string('start', None, 'Start date formatted as year/month/day')
gflags.DEFINE_string('end', None, 'End date (inclusive) formatted as year/month/day')
gflags.DEFINE_multistring('filter', [], 'Add filters')
gflags.DEFINE_multistring('exclude', [], 'Add exclusions')
gflags.DEFINE_boolean(
    'details',
    True,
    'If set to False, only the id# of the start and ending stations will be provided')

compose_mult = partial(reduce, compose)

STATION_DETAILS = set(
    ('start station longitude',
     'start station latitude',
     'end station longitude',
     'end station latitude',
     'end station name',
     'start station name'))

def int_or_none(value):
    try:
        return int(value)
    except:
        return None

def parse_date(d):
    if d is None:
        return None
    y, m, d = map(int, d.split('-'))
    return date(y, m, d)

def validate_date(d):
    try:
        parse_date(d)
        return True
    except:
        return False


gflags.RegisterValidator('start',
                        validate_date,
                        message='Incorrect date format',
                        flag_values=FLAGS)

gflags.RegisterValidator('end',
                        validate_date,
                        message='Incorrect date format',
                        flag_values=FLAGS)


def remove_station_details(record):
    """Removes station details from a record"""
    retval = {}
    for key, value in record.items():
        if key in STATION_DETAILS:
            continue
        retval[key] = value
    return retval


class DateChecker():
    last_str = None
    last_date = None
    def __init__(self, start, stop):
        self.start, self.stop = map(parse_date, (start, stop))

        if None not in (self.start, self.stop):
            self.start, self.stop = sorted((self.start, self.stop))

        self.start_file = (self.start.year, self.start.month) if self.start else None
        self.stop_file = (self.stop.year, self.stop.month) if self.stop else None
        
    @staticmethod
    def parse_filename(fn):
        fn = basename(fn)
        fn = fn.split()[0].split('-')
        return tuple(map(int, fn))

    @classmethod
    def parse_date_from_timestamp(cls, d):
        d = d.split()[0]
        if cls.last_str == d.split()[0]:
            return cls.last_date
        cls.last_str = d
        d = date(*map(int, d.split('-')))
        cls.last_date = d
        return d

    def valid_row(self, row):
        date = self.parse_date_from_timestamp(row['starttime'])
        if self.stop and date > self.stop:
            raise StopIteration
        return self.start is None or date >= self.start
        

    def valid_file(self, fn):
        fn = self.parse_filename(fn)
        year, month = fn
        if self.stop_file is not None:
            if (year,month) > self.stop_file:
                return False
        if self.start_file is not None:
            if (year, month) < self.stop_file:
                return False
        return True




def rides(files=[]):
    def float(obj, _cache={}, float=__builtins__['float']):

        date_checker = DateChecker(FLAGS.start, FLAGS.end)
        start, end = map(parse_date, (FLAGS.start, FLAGS.end))
        if None not in (start,end):
            start,end = sorted((start, end))

        # Its important we have a memoized version of float to be
        # absolutely sure two strings with the same source compare
        # exactly.  This isn't defind in the global scope for a few
        # reasons - for one, we only really care about comparisons
        # within a single pass for the benefit of
        # citibike.station:unique_stations's ability to dedup
        # statiions and not accidentally raise a ValueError.  So
        # there's no point to corrupting the global scope, interning
        # extra floats or keeping any of the interned relationships
        # after rides ends.

        if obj in _cache:
            return _cache[obj]

        f = float(obj)
        _cache[obj] = f
        return f

    date_checker = DateChecker(FLAGS.start, FLAGS.end)
    files = files or sorted(glob(join(FLAGS.cache, '*.csv.bz2')))
    files = filter(date_checker.valid_file, files)
    if not files:
        return
    for ride in chain(*(DictReader(BZ2File(f)) for f in files)):
        if not date_checker.valid_row(ride): continue 
        ride['start station id'] = int_or_none(ride['start station id'])
        ride['end station id'] = int_or_none(ride['end station id'])
        ride['tripduration'] = int_or_none(ride['tripduration'])
        ride['bikeid'] = int_or_none(ride['bikeid'])
        ride['gender'] = int_or_none(ride['gender'])
        ride['birth year'] = int_or_none(ride['birth year'])
        ride['start station latitude'] = float(ride['start station latitude'])
        ride['start station longitude'] = float(ride['start station longitude'])
        ride['end station latitude'] = float(ride['end station latitude'])
        ride['end station longitude'] = float(ride['end station longitude'])
        yield ride


def run():
    qs = Query(rides())
    qs = qs.filter(**dict(f.split('=') for f in FLAGS.filter))
    qs = qs.exclude(**dict(f.split('=') for f in FLAGS.exclude))

    for ride in qs:
        if not FLAGS.details:
            ride = ride(remove_station_details(ride))
        try:
            print(dumps(ride))
        except IOError:
            break


class Query:

    def __init__(self, seq):
        self.seq = seq
        self.filter_func = []

    def filter(self, **kwargs):
        q = Query(self)
        q.__set_filter(kwargs)
        return q

    def exclude(self, **kwargs):
        q = Query(self)
        q.__set_exclude(kwargs)
        return q

    def __set_filter(self, kwargs):
        for key, value in kwargs.items():
            def a_filter(seq, key=key, value=value):
                if True:
                    for item in seq:
                        if item[key] == value:
                            yield item
            self.filter_func.append(a_filter)

    def __set_exclude(self, kwargs):
        for key, value in kwargs.items():
            def a_filter(seq, key=key, value=value):
                if True:
                    for item in seq:
                        if item[key] != value:
                            yield item
            self.filter_func.append(a_filter)

    def __iter__(self):
        if self.filter_func:
            seq = self.seq
            for f in self.filter_func:
                seq = f(seq)
            for i in seq:
                yield i
        else:
            for x in self.seq:
                yield x


def main(argv):
    try:
        argv = FLAGS(argv)[1:]
    except (gflags.FlagsError, KeyError, IndexError) as e:
        sys.stderr.write("%s\nUsage: %s \n%s\n" % (
            e, os.path.basename(sys.argv[0]), FLAGS))
        return 1
    try:
        run()
    except IOError:
        return 0
    finally:
        sys.stderr.close()
        sys.stdout.close()



from citibike.station import unique_stations
from itertools import imap
from gflags import FLAGS
from sys import stdin
from geojson import LineString, Point, GeometryCollection, dumps


def station_ids_to_stations(station_ids):    
    stations = unique_stations(subset=station_ids)
    stations = dict((station['id'], station) for station in stations)
    
    for station_id in station_ids:
        yield stations.get(station_id)

def stations_to_geojson(stations):
    stations = list(stations)
    items = [
        LineString([(station['longitude'], station['latitude']) for station in stations])
        ]
    for station in stations:
        items.append(Point(coordinates=(
            station['longitude'],
            station['latitude'],
        ),
                           properties={'name':station['name'],
                                       'id': station['id']}))
    return GeometryCollection(items)


def run(station_id_stream):
    station_ids = filter(None, station_id_stream.read().split())
    station_ids = map(int, station_ids)
    stations = station_ids_to_stations(station_ids)
    print dumps(stations_to_geojson(stations))


def main(argv): 
    try:
        argv = FLAGS(argv)[1:]
        if len(argv) not in (0, 1):
            raise ValueError("Too many parameters - only accepts a single optional parameter")
    except (gflags.FlagsError, KeyError, IndexError, ValueError) as e:
        stderr.write("%s\nUsage: %s [infile.txt]\n%s\nAccepts a list of station_ids" % (
            e, basename(argv[0]), FLAGS))
        return 1
    if argv:
        station_id_stream = open(argv[0])
    else:
        station_id_stream = stdin
    return run(station_id_stream)



#!/usr/bin/env python3

from bz2file import BZ2File
from citibike.cache import cache_filename
from citibike.reader import rides
from csv import DictReader
from functional import compose, partial
from functools import reduce
from gflags import FLAGS
from glob import glob
from itertools import chain
from json import load, dump, dumps
from os import environ
from os.path import basename, exists, join
from sys import argv, stderr


def get_stations():
    for ride in rides():
        yield ride['start station id'], {'longitude': ride['start station longitude'],
                                         'latitude': ride['start station latitude'],
                                         'name': ride['start station name'],
                                         'id': ride['start station id'],
                                         }
        yield ride['end station id'], {'longitude': ride['end station longitude'],
                                       'latitude': ride['end station latitude'],
                                       'name': ride['end station name'],
                                       'id': ride['end station id'],
                                       }


def unique_stations(subset=None):
    """Yield a dictionary for each unique station.
    
    Args:
      subset: An optional set of station_ids to search for.  The
              search will end once all of these will be identified.

    Yields:
      Dictionaries with station details. 
      {'longitude': ...,
       'latitude': ...,
       'name': ...,
       'id': ...}
    """
    subset = set(subset)
    stations = {}
    for station_id, station_data in get_stations():
        if subset is not None:
            if not subset:
                break
            if station_id not in subset:
                continue 
            subset.remove(station_id)
        previous_station_data = stations.get(station_id)
        if previous_station_data is not None:
            if previous_station_data == station_data:
                continue
            raise ValueError(
                "Inconsistent station data for station #%d: %r != %r" %
                (station_id, previous_station_data, station_data))
        else:
            stations[station_id] = station_data
    for station_data in stations.values():
        yield station_data


def run():
    for station_data in unique_stations():
        print(dumps(station_data))


def main(argv):
    try:
        argv = FLAGS(argv)[1:]
    except (gflags.FlagsError, KeyError, IndexError) as e:
        stderr.write("%s\nUsage: %s \n%s\n" % (
            e, basename(argv[0]), FLAGS))
        return 1
    return run()



import gflags
from os import environ
from os.path import join

FLAGS = gflags.FLAGS


gflags.DEFINE_string(
    'cache',
    "%(HOME)s/.citibike/cache" %
    environ,
    'Location to write downloaded files.  Defaults to $HOME/.citibike/cache')


def cache_filename(name):
    return join(FLAGS.cache, name)



"""Download original citibike data.

This application downloads the original citibike data and reencodes it
from a zip to a bz2 file."""

from pipes import quote
from tempfile import TemporaryFile
from distutils.dir_util import mkpath
from os import environ, makedirs, unlink
from os.path import join, exists
import errno
import requests
from bz2file import BZ2File
from zipfile import ZipFile
import gflags
from progressbar import ProgressBar
import citibike.reader

FLAGS = gflags.FLAGS

gflags.DEFINE_integer('year', 2013, 'Year of first file to download')
gflags.DEFINE_integer('month', 7, 'Month of first file to download')
gflags.DEFINE_string(
    'filename',
    '%(year)04d%(month)02d-citibike-tripdata.zip',
    'Template for building citibike filename/url')
gflags.DEFINE_string(
    'url',
    'https://s3.amazonaws.com/tripdata/%(filename)s',
    'URL template from which to download')
gflags.DEFINE_string(
    'csv',
    '%(year)04d-%(month)02d - Citi Bike trip data.csv',
    'Name of CSV file in zip file')
gflags.DEFINE_string(
    'compressed_csv',
    '%(year)04d-%(month)02d - Citi Bike trip data.csv.bz2',
    'Name of bz2 compressed CSV file')
gflags.DEFINE_string('chunk_size', 2 ** 20, 'Download chunk size')


def month_counter(year=None, month=None):
    year = year or FLAGS.year
    month = month or FLAGS.month
    while True:
        while month <= 12:
            yield year, month
            month += 1
        month = 1
        year += 1


def path_csv_and_urls(year=None, month=None, filename=None, url=None):
    for year, month in month_counter(year, month):
        filename = FLAGS.filename % vars()
        url = FLAGS.url % vars()
        path = join(FLAGS.cache, FLAGS.compressed_csv % vars())
        csv = FLAGS.csv % vars()
        yield path, csv, url


def download(target=None):

    target = target or FLAGS.cache

    mkpath(target)
    print "Starting download"
    for path, csv, url in path_csv_and_urls():
        if exists(path):
            print("Skipping %s, %s exists" % (url, path))
            continue
        req = requests.get(url, stream=True)
        if req.status_code != 200:
            return req.status_code
        print("Downloading: %(url)s " % vars())
        with TemporaryFile() as tf:
            l = 0
            size = int(req.headers['content-length'])
            pb = ProgressBar(size).start()
            for chunk in req.iter_content(FLAGS.chunk_size):
                l += len(chunk)
                pb.update(l)
                tf.write(chunk)
            tf.flush()
            tf.seek(0)
            pb.finish()
            print("Recompressing to: " + quote(path))
            size = (
                i.file_size for i in ZipFile(tf).infolist() if i.filename == csv).next()
            l = 0
            try:
                with ZipFile(tf).open(csv) as csv_file:
                    pb = ProgressBar(size).start()
                    with BZ2File(path, "w") as target_file:
                        while True:
                            chunk = csv_file.read(FLAGS.chunk_size)
                            if not chunk:
                                break
                            l += len(chunk)
                            pb.update(l)
                            target_file.write(chunk)
            except KeyboardInterrupt:
                unlink(path)
                break
            finally:
                pb.finish()


def main(argv):
    try:
        argv = FLAGS(argv)[1:]
    except (gflags.FlagsError, KeyError, IndexError) as e:
        sys.stderr.write("%s\nUsage: %s \n%s\n" % (
            e, os.path.basename(sys.argv[0]), FLAGS))
        return 1

    download()



#!/usr/bin/env python
"""
py-citibike
"""

from setuptools import setup

setup(
    name='citibike',
    version='0.0.4',
    author='Adam DePrince',
    author_email='deprince@googlealumni.com',
    description='',
    long_description=__doc__,
    py_modules=[
        'citibike/__init__',
        'citibike/reader',
        'citibike/download',
        'citibike/station/__init__',
        'citibike/station/mapper',
    ],
    packages=['citibike'],
    zip_safe=True,
    include_package_data=True,
    classifiers=[
    ],
    scripts=[
        # 'scripts/citibike_to_postgresql'
        'scripts/citibike_download_originals',
        # Downloads and unpacks original citibike data
        'scripts/citibike_dump',
        # Downloads and unpacks original citibike data
        'scripts/citibike_stations',
        # Dumps station data and manages station data cache
        'scripts/citibike_station_mapper'
    ],
    url='https://github.com/adamdeprince/python-citibike-data',
    install_requires=[
        'bz2file',
        'functional',
        'geojson',
        'python-gflags',
        'requests',
        'progressbar',
    ]
)
